import dash

from webviz_config.common_cache import CACHE
from webviz_config.containers import _example_container


def test_example_container(dash_duo):

    app = dash.Dash(__name__)
    app.config.suppress_callback_exceptions = True
    CACHE.init_app(app.server)
    title = "Example"
    page = _example_container.ExampleContainer(app, title)
    app.layout = page.layout
    dash_duo.start_server(app)
    btn = dash_duo.find_element(f"#{page.button_id}")
    assert btn.text == "Submit"
    text = dash_duo.find_element(f"#{page.div_id}")
    assert text.text == "Button has been pressed 0 times."
    btn.click()
    dash_duo.wait_for_contains_text(
        f"#{page.div_id}", "Button has been pressed 1 times", timeout=2
    )
    assert text.text == "Button has been pressed 1 times."



import io
import abc
import base64
import zipfile
from uuid import uuid4

import bleach
from dash.dependencies import Input, Output
import webviz_core_components as wcc


class WebvizContainerABC(abc.ABC):
    """All webviz containers need to subclass this abstract base class,
    e.g.

    ```python
    class MyContainer(WebvizContainerABC):

        def __init__(self):
            ...

        def layout(self):
            ...
    ```
    """

    # This is the default set of buttons to show in the rendered container
    # toolbar. If the list is empty, the subclass container layout will be
    # used directly, without any visual encapsulation layout from this
    # abstract base class. The containers subclassing this abstract base class
    # can override this variable setting by defining a class constant with
    # the same name.
    #
    # Some buttons will only appear if in addition necessary data is available.
    # E.g. download of zip archive will only appear if the container also
    # has defined the corresponding callback, and contact person will only
    # appear if the user configuration file has this information.
    TOOLBAR_BUTTONS = [
        "screenshot",
        "expand",
        "download_zip",
        "contact_person",
        "guided_tour",
    ]

    # List of container specific assets which should be copied
    # over to the ./assets folder in the generated webviz app.
    # This is typically custom JavaScript and/or CSS files.
    # All paths in the returned ASSETS list should be absolute.
    ASSETS = []

    @property
    @abc.abstractmethod
    def layout(self):
        """This is the only required function of a Webviz Container.
        It returns a Dash layout which by webviz-config is added to
        the main Webviz application.
        """

    @property
    def _container_wrapper_id(self):
        # pylint: disable=attribute-defined-outside-init
        # We do not have a __init__ method in this abstract base class
        if not hasattr(self, "_container_wrapper_uuid"):
            self._container_wrapper_uuid = uuid4()
        return f"container-wrapper-{self._container_wrapper_uuid}"

    @property
    def container_data_output(self):
        # pylint: disable=attribute-defined-outside-init
        # We do not have a __init__ method in this abstract base class
        self._add_download_button = True
        return Output(self._container_wrapper_id, "zip_base64")

    @property
    def container_data_requested(self):
        return Input(self._container_wrapper_id, "data_requested")

    @staticmethod
    def _reformat_tour_steps(steps):
        return [
            {"selector": "#" + step["id"], "content": step["content"]} for step in steps
        ]

    @staticmethod
    def container_data_compress(content):
        byte_io = io.BytesIO()

        with zipfile.ZipFile(byte_io, "w") as zipped_data:
            for data in content:
                zipped_data.writestr(data["filename"], data["content"])

        byte_io.seek(0)

        return base64.b64encode(byte_io.read()).decode("ascii")

    def container_layout(self, contact_person=None):
        """This function returns (if the class constant SHOW_TOOLBAR is True,
        the container layout wrapped into a common webviz config container
        component, which provides some useful buttons like download of data,
        show data contact person and download container content to png.

        CSV download button will only appear if the container class a property
        `csv_string` which should return the appropriate csv data as a string.

        If TOOLBAR_BUTTONS is empty, this functions returns the same
        dash layout as the container class provides directly.
        """

        buttons = self.__class__.TOOLBAR_BUTTONS.copy()

        if contact_person is None:
            contact_person = {}
        else:
            # Sanitize the configuration user input
            for key in contact_person:
                contact_person[key] = bleach.clean(str(contact_person[key]))

        if "download_zip" in buttons and not hasattr(self, "_add_download_button"):
            buttons.remove("download_zip")

        if buttons:
            return wcc.WebvizContainerPlaceholder(
                id=self._container_wrapper_id,
                buttons=buttons,
                contact_person=contact_person,
                children=[self.layout],
                tour_steps=WebvizContainerABC._reformat_tour_steps(
                    self.tour_steps  # pylint: disable=no-member
                )
                if "guided_tour" in buttons and hasattr(self, "tour_steps")
                else [],
            )
        return self.layout



from pkg_resources import get_distribution, DistributionNotFound

from ._localhost_token import LocalhostToken
from ._localhost_open_browser import LocalhostOpenBrowser
from ._localhost_certificate import LocalhostCertificate
from ._is_reload_process import is_reload_process
from ._container_abc import WebvizContainerABC
from ._theme_class import WebvizConfigTheme

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



class WebvizConfigTheme:
    """Webviz config themes are all instances of this class. The only mandatory
    property is the theme name set at initialization.
    """

    def __init__(self, theme_name):
        self.theme_name = theme_name

        self._csp = {
            "default-src": "'none'",
            "connect-src": "'self'",
            "prefetch-src": "'self'",
            "style-src": ["'self'", "'unsafe-inline'"],  # [1]
            "script-src": [
                "'self'",
                "'sha256-jZlsGVOhUAIcH+4PVs7QuGZkthRMgvT2n0ilH6/zTM0='",
            ],  # [2]
            "img-src": ["'self'", "data:"],
            "navigate-to": "'self'",
            "base-uri": "'self'",
            "form-action": "'self'",
            "frame-ancestors": "'none'",
            "child-src": "'none'",
            "object-src": "'self'",
            "plugin-types": "application/pdf",
        }

        """
        These are the current exceptions to the most strict CSP setup:
            [1] unsafe-inline for style still needed by plotly
                (https://github.com/plotly/plotly.js/issues/2355)
            [2] https://github.com/plotly/dash/issues/630
        """

        self._feature_policy = {
            "camera": "'none'",
            "geolocation": "'none'",
            "microphone": "'none'",
            "payment": "'none'",
        }

        self._external_stylesheets = []
        self._assets = []
        self._plotly_layout = {}

    def adjust_csp(self, dictionary, append=True):
        """If the default CSP settings needs to be changed, this function can
        be called by giving in a dictionary with key-value pairs which should
        be changed. If `append=True`, the CSP sources given in the dictionary
        are appended to the whitelisted sources in the default configuration.
        If not the input value source list is instead replacing the default
        whitelisted sources.
        """

        for key, value in dictionary.items():
            if append and key in self._csp:
                self._csp[key] += value
            else:
                self._csp[key] = value

    @property
    def csp(self):
        """Returns the content security policy settings for the theme."""
        return self._csp

    @property
    def feature_policy(self):
        """Returns the feature policy settings for the theme."""
        return self._feature_policy

    @property
    def plotly_layout(self):
        return self._plotly_layout

    @plotly_layout.setter
    def plotly_layout(self, plotly_layout):
        """Layout object of Plotly graph objects."""
        self._plotly_layout = plotly_layout

    @property
    def external_stylesheets(self):
        return self._external_stylesheets

    @external_stylesheets.setter
    def external_stylesheets(self, external_stylesheets):
        """Set optional external stylesheets to be used in the Dash
        application. The input variable `external_stylesheets` should be
        a list."""
        self._external_stylesheets = external_stylesheets

    @property
    def assets(self):
        return self._assets

    @assets.setter
    def assets(self, assets):
        """Set optional theme assets to be copied over to the `./assets` folder
        when the webviz dash application is created. The input variable
        `assets` should be a list of absolute file paths to the different
        assets.
        """
        self._assets = assets



from uuid import uuid4

import dash_html_components as html
from dash.dependencies import Input, Output

from .. import WebvizContainerABC


class ExampleContainer(WebvizContainerABC):
    def __init__(self, app, title: str):
        self.title = title

        self.button_id = f"submit-button-{uuid4()}"
        self.div_id = f"output-state-{uuid4()}"

        self.set_callbacks(app)

    @property
    def layout(self):
        return html.Div(
            [
                html.H1(self.title),
                html.Button(id=self.button_id, n_clicks=0, children="Submit"),
                html.Div(id=self.div_id),
            ]
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.div_id, "children"), [Input(self.button_id, "n_clicks")]
        )
        def _update_output(n_clicks):
            return "Button has been pressed {} times.".format(n_clicks)



import os
import re
import sys
import pathlib
import getpass
import datetime
import subprocess  # nosec

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa

from .utils import terminal_colors


NAME = x509.Name(
    [
        x509.NameAttribute(NameOID.COUNTRY_NAME, "NO"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Trondheim"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Webviz"),
        x509.NameAttribute(NameOID.COMMON_NAME, f"{getpass.getuser()}"),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Webviz"),
    ]
)

CA_KEY_FILENAME = "ca.key"
CA_CRT_FILENAME = "ca.crt"

SERVER_KEY_FILENAME = "server.key"
SERVER_CRT_FILENAME = "server.crt"


def user_data_dir():
    """Returns platform specific path to store user application data
    """

    if sys.platform == "win32":
        return os.path.normpath(os.path.expanduser("~/Application Data/webviz"))

    if sys.platform == "darwin":
        return os.path.expanduser("~/Library/Application Support/webviz")

    return os.path.expanduser("~/.local/share/webviz")


def create_key(key_path):

    key = rsa.generate_private_key(
        public_exponent=65537, key_size=2048, backend=default_backend()
    )

    with open(key_path, "wb") as filehandle:
        filehandle.write(
            key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    return key


def certificate_template(subject, issuer, public_key, certauthority=False):

    if certauthority:
        not_valid_after = datetime.datetime.utcnow() + datetime.timedelta(days=365 * 10)

    else:  # shorter valid length for on-the-fly certificates
        not_valid_after = datetime.datetime.utcnow() + datetime.timedelta(days=7)

    return (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(not_valid_after)
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("localhost")]), critical=True
        )
        .add_extension(
            x509.BasicConstraints(ca=certauthority, path_length=None), critical=True
        )
    )


def create_ca(args):

    directory = user_data_dir()

    os.makedirs(directory, exist_ok=True)

    ca_key_path = os.path.join(directory, CA_KEY_FILENAME)
    ca_crt_path = os.path.join(directory, CA_CRT_FILENAME)

    if not args.force and os.path.isfile(ca_crt_path):
        raise OSError(
            f"The file {ca_crt_path} already exists. Add the "
            "command line flag --force if you want to overwrite"
        )

    key = create_key(ca_key_path)

    subject = issuer = NAME

    cert = certificate_template(
        subject, issuer, key.public_key(), certauthority=True
    ).sign(key, hashes.SHA256(), default_backend())

    with open(ca_crt_path, "wb") as filehandle:
        filehandle.write(cert.public_bytes(serialization.Encoding.PEM))

    # The SHA1 hash here is used only for the user to be able to compare with
    # the SHA1 hash generated by Chrome for visual comparison/comfort by the user.
    sha1 = "-".join(
        re.findall(".{8,8}", cert.fingerprint(hashes.SHA1()).hex())  # nosec
    ).upper()

    installed = False
    if args.auto_install:
        try:
            subprocess.run(  # nosec
                [
                    "certutil",
                    "-d",
                    f"sql:{pathlib.Path.home() / '.pki' / 'nssdb'}",
                    "-A",
                    "-t",
                    "CT,C,c",
                    "-n",
                    "webviz",
                    "-i",
                    ca_crt_path,
                ],
                check=True,
            )
            installed = True
            print(
                f"{terminal_colors.GREEN}{terminal_colors.BOLD}"
                "Successfully installed webviz certificate. "
                "Ready to browse applications on localhost."
                f"{terminal_colors.END}"
            )
        except (PermissionError, subprocess.CalledProcessError):
            print(
                f"{terminal_colors.RED}"
                "Automatic installation of webviz certificate failed. "
                "Falling back to manual installation."
                f"{terminal_colors.END}"
            )

    if not installed:
        print(
            f"""{terminal_colors.BLUE}{terminal_colors.BOLD}
 Created CA key and certificate files (both saved in {directory}).
 Keep the key file ({CA_KEY_FILENAME}) private. The certificate file
 ({CA_CRT_FILENAME}) is not sensitive, and you can import it in
 your browser(s).

 To install it in Chrome:

    - Open Chrome and go to chrome://settings/privacy
    - Select "Manage certificates"
    - Under the tab "Trusted Root Certificatation Authorities", click "Import"
    - Go to {directory} and select the created certificate ({CA_CRT_FILENAME}).
    - Click "Next" and select "Place all certificates in the following store"
    - Click "Next" and then "Finished"
    - If a dialog box appears, you can verify that the displayed thumbprint is
      the same as this one:
       {sha1}
    - Restart Chrome

 When done, you do not have to rerun "webviz certificate" or do this procedure
 before the certificate expiry date has passed. The certificate is only valid
 for localhost.{terminal_colors.END}"""
        )


def create_certificate(directory):

    ca_directory = user_data_dir()
    ca_key_path = os.path.join(ca_directory, CA_KEY_FILENAME)
    ca_crt_path = os.path.join(ca_directory, CA_CRT_FILENAME)

    server_key_path = os.path.join(directory, SERVER_KEY_FILENAME)
    server_crt_path = os.path.join(directory, SERVER_CRT_FILENAME)

    if not os.path.isfile(ca_key_path) or not os.path.isfile(ca_crt_path):
        raise RuntimeError(
            "Could not find CA key and certificate. Please "
            'run the command "webviz certificate --auto-install" and '
            "try again"
        )

    with open(ca_key_path, "rb") as filehandle:
        ca_key = serialization.load_pem_private_key(
            data=filehandle.read(), password=None, backend=default_backend()
        )

    with open(ca_crt_path, "rb") as filehandle:
        ca_crt = x509.load_pem_x509_certificate(
            data=filehandle.read(), backend=default_backend()
        )

    server_key = create_key(server_key_path)

    crt = (
        certificate_template(NAME, ca_crt.subject, server_key.public_key())
        .add_extension(
            critical=True,
            extension=x509.KeyUsage(
                digital_signature=True,
                key_encipherment=True,
                content_commitment=True,
                data_encipherment=False,
                key_agreement=False,
                encipher_only=False,
                decipher_only=False,
                key_cert_sign=False,
                crl_sign=False,
            ),
        )
        .add_extension(
            critical=False,
            extension=x509.AuthorityKeyIdentifier.from_issuer_public_key(
                ca_key.public_key()
            ),
        )
        .sign(private_key=ca_key, algorithm=hashes.SHA256(), backend=default_backend())
    )

    with open(server_crt_path, "wb") as filehandle:
        filehandle.write(crt.public_bytes(encoding=serialization.Encoding.PEM))



from uuid import uuid4
from pathlib import Path
from collections import OrderedDict

import numpy as np
import pandas as pd
import plotly.express as px
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output
import webviz_core_components as wcc

from .. import WebvizContainerABC
from ..webviz_store import webvizstore
from ..common_cache import CACHE


class TablePlotter(WebvizContainerABC):
    """### TablePlotter

This container adds a plotter to the webviz instance, using tabular data from
a provided csv file. If feature is requested, the data could also come from
a database.

* `csv_file`: Path to the csv file containing the tabular data. Either absolute
              path or relative to the configuration file.
* `plot_options`: A dictionary of plot options to initialize the plot with
* `lock`: If `True`, only the plot is shown, all dropdowns for changing
          plot options are hidden.
"""

    def __init__(
        self,
        app,
        csv_file: Path,
        plot_options: dict = None,
        filter_cols: list = None,
        lock: bool = False,
    ):

        self.plot_options = plot_options if plot_options else {}
        self.graph_id = f"graph-id{uuid4()}"
        self.lock = lock
        self.csv_file = csv_file
        self.data = get_data(self.csv_file)
        self.set_filters(filter_cols)
        self.columns = list(self.data.columns)
        self.numeric_columns = list(
            self.data.select_dtypes(include=[np.number]).columns
        )
        self.selector_row = f"selector-row{uuid4()}"
        self.plot_option_id = f"plot-option{uuid4()}"
        self.set_callbacks(app)

    def set_filters(self, filter_cols):
        self.filter_cols = []
        self.filter_ids = {}
        self.use_filter = False
        if filter_cols:
            for col in filter_cols:
                if col in self.data.columns:
                    if self.data[col].nunique() != 1:
                        self.filter_cols.append(col)
            if self.filter_cols:
                self.use_filter = True
                self.filter_ids = {
                    col: f"{col}-{str(uuid4())}" for col in self.filter_cols
                }

    def add_webvizstore(self):
        return [(get_data, [{"csv_file": self.csv_file}])]

    @property
    def plots(self):
        """A list of available plots and their options"""
        return {
            "scatter": ["x", "y", "size", "color", "facet_col"],
            "histogram": [
                "x",
                "y",
                "color",
                "facet_col",
                "barmode",
                "barnorm",
                "histnorm",
            ],
            "bar": ["x", "y", "color", "facet_col"],
            "scatter_3d": ["x", "y", "z", "size", "color"],
            "line": ["x", "y", "color", "line_group", "facet_col"],
            "line_3d": ["x", "y", "z", "color"],
            "box": ["x", "y", "color", "facet_col"],
            "violin": ["x", "y", "color", "facet_col"],
            "scatter_matrix": ["dimensions", "size", "color"],
            "parallel_coordinates": ["dimensions"],
            "parallel_categories": ["dimensions"],
            "density_contour": ["x", "y", "color", "facet_col"],
        }

    @property
    def plot_args(self):
        """A list of possible plot options and their default values"""
        return OrderedDict(
            {
                "x": {
                    "options": self.columns,
                    "value": self.plot_options.get("x", self.columns[0]),
                    "multi": False,
                },
                "y": {
                    "options": self.columns,
                    "value": self.plot_options.get("y", self.columns[0]),
                    "multi": False,
                },
                "z": {
                    "options": self.columns,
                    "value": self.plot_options.get("z", self.columns[0]),
                    "multi": False,
                },
                "size": {
                    "options": self.numeric_columns,
                    "value": self.plot_options.get("size", None),
                    "multi": False,
                },
                "color": {
                    "options": self.columns,
                    "value": self.plot_options.get("color", None),
                    "multi": False,
                },
                "facet_col": {
                    "options": self.columns,
                    "value": self.plot_options.get("facet_col", None),
                    "multi": False,
                },
                "line_group": {
                    "options": self.columns,
                    "value": self.plot_options.get("line_group", None),
                    "multi": False,
                },
                "barmode": {
                    "options": ["stack", "group", "overlay", "relative"],
                    "value": self.plot_options.get("barmode", "stack"),
                    "multi": False,
                },
                "barnorm": {
                    "options": ["fraction", "percent"],
                    "value": self.plot_options.get("barnorm", None),
                    "multi": False,
                },
                "histnorm": {
                    "options": [
                        "percent",
                        "propability",
                        "density",
                        "propability density",
                    ],
                    "value": self.plot_options.get("histnorm", None),
                    "multi": False,
                },
                "trendline": {
                    "options": self.numeric_columns,
                    "value": None,
                    "multi": False,
                },
                "dimensions": {
                    "options": self.columns,
                    "value": self.plot_options.get("dimensions", self.columns),
                    "multi": True,
                },
            }
        )

    def filter_layout(self):
        """Makes dropdowns for each dataframe column used for filtering."""
        if not self.use_filter:
            return None
        df = self.data
        dropdowns = [html.H4("Set filters")]
        for col in self.filter_cols:
            if df[col].dtype == np.float64 or df[col].dtype == np.int64:
                min_val = df[col].min()
                max_val = df[col].max()
                mean_val = df[col].mean()
                dropdowns.append(
                    html.Div(
                        children=[
                            html.Details(
                                open=True,
                                children=[
                                    html.Summary(col.lower().capitalize()),
                                    dcc.RangeSlider(
                                        id=self.filter_ids[col],
                                        min=min_val,
                                        max=max_val,
                                        step=(max_val - min_val) / 10,
                                        marks={
                                            min_val: f"{min_val:.2f}",
                                            mean_val: f"{mean_val:.2f}",
                                            max_val: f"{max_val:.2f}",
                                        },
                                        value=[min_val, max_val],
                                    ),
                                ],
                            )
                        ]
                    )
                )
            else:
                elements = list(self.data[col].unique())
                dropdowns.append(
                    html.Div(
                        children=[
                            html.Details(
                                open=True,
                                children=[
                                    html.Summary(col.lower().capitalize()),
                                    dcc.Dropdown(
                                        id=self.filter_ids[col],
                                        options=[
                                            {"label": i, "value": i} for i in elements
                                        ],
                                        value=elements,
                                        multi=True,
                                    ),
                                ],
                            )
                        ]
                    )
                )
        return dropdowns

    def plot_option_layout(self):
        """Renders a dropdown widget for each plot option"""
        divs = []
        # The plot type dropdown is handled separate
        divs.append(
            html.Div(
                style=self.style_options_div,
                children=[
                    html.H4("Set plot options"),
                    html.P("Plot type"),
                    dcc.Dropdown(
                        id=f"{self.plot_option_id}-plottype",
                        clearable=False,
                        options=[{"label": i, "value": i} for i in self.plots],
                        value=self.plot_options.get("type", "scatter"),
                    ),
                ],
            )
        )
        # Looping through all available plot options
        # and renders a dropdown widget
        for key, arg in self.plot_args.items():
            divs.append(
                html.Div(
                    style=self.style_options_div,
                    id=f"{self.plot_option_id}-div-{key}",
                    children=[
                        html.P(key),
                        dcc.Dropdown(
                            id=f"{self.plot_option_id}-{key}",
                            clearable=False,
                            options=[{"label": i, "value": i} for i in arg["options"]],
                            value=arg["value"],
                            multi=arg["multi"],
                        ),
                    ],
                )
            )
        return divs

    @property
    def style_options_div(self):
        """Style for active plot options"""
        return {"display": "grid"}

    @property
    def style_options_div_hidden(self):
        """Style for hidden plot options"""
        return {"display": "none"}

    @property
    def style_page_layout(self):
        """Simple grid layout for the page"""
        if self.lock:
            return {}
        return {
            "display": "grid",
            "align-content": "space-around",
            "justify-content": "space-between",
            "grid-template-columns": "1fr 5fr 1fr" if self.use_filter else "1fr 5fr",
        }

    @property
    def style_selectors(self):
        return {"display": "none"} if self.lock else {}

    @property
    def layout(self):
        return html.Div(
            children=[
                html.Div(
                    style=self.style_page_layout,
                    children=[
                        html.Div(
                            id=self.selector_row,
                            style=self.style_selectors,
                            children=self.plot_option_layout(),
                        ),
                        html.Div(
                            style={"height": "100%"},
                            children=wcc.Graph(id=self.graph_id),
                        ),
                        html.Div(children=self.filter_layout()),
                    ],
                )
            ]
        )

    @property
    def plot_output_callbacks(self):
        """Creates list of output dependencies for callback
        The outputs are the graph, and the style of the plot options"""
        outputs = []
        outputs.append(Output(self.graph_id, "figure"))
        for plot_arg in self.plot_args.keys():
            outputs.append(Output(f"{self.plot_option_id}-div-{plot_arg}", "style"))
        return outputs

    @property
    def plot_input_callbacks(self):
        """Creates list of input dependencies for callback
        The inputs are the plot type and the current value
        for each plot option
        """
        inputs = []
        inputs.append(Input(f"{self.plot_option_id}-plottype", "value"))
        for plot_arg in self.plot_args.keys():
            inputs.append(Input(f"{self.plot_option_id}-{plot_arg}", "value"))
        for filtcol in self.filter_cols:
            inputs.append(Input(self.filter_ids[filtcol], "value"))
        return inputs

    def set_callbacks(self, app):
        @app.callback(self.container_data_output, [self.container_data_requested])
        def _user_download_data(data_requested):
            return (
                WebvizContainerABC.container_data_compress(
                    [
                        {
                            "filename": "table_plotter.csv",
                            "content": get_data(self.csv_file).to_csv(),
                        }
                    ]
                )
                if data_requested
                else ""
            )

        @app.callback(self.plot_output_callbacks, self.plot_input_callbacks)
        def _update_output(*args):
            """Updates the graph and shows/hides plot options"""
            plot_type = args[0]
            # pylint: disable=protected-access
            plotfunc = getattr(px._chart_types, plot_type)
            plotargs = {}
            div_style = []
            data = self.data
            # Filter dataframe if filter columns are available
            if self.use_filter:
                plot_inputs = args[1 : -len(self.filter_cols)]
                filter_inputs = args[-len(self.filter_cols) :]
                data = filter_dataframe(data, self.filter_cols, filter_inputs)
            else:
                plot_inputs = args[1:]
            for name, plot_arg in zip(self.plot_args.keys(), plot_inputs):
                if name in self.plots[plot_type]:
                    plotargs[name] = plot_arg
                    div_style.append(self.style_options_div)
                else:
                    div_style.append(self.style_options_div_hidden)

            return (plotfunc(data, **plotargs), *div_style)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def get_data(csv_file) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_dataframe(dframe, columns, column_values):
    df = dframe.copy()
    if not isinstance(columns, list):
        columns = [columns]
    for filt, col in zip(column_values, columns):
        if isinstance(filt, list):
            if df[col].dtype == np.float64 or df[col].dtype == np.int64:
                df = df.loc[df[col].between(filt[0], filt[1])]
            else:
                df = df.loc[df[col].isin(filt)]
        else:
            df = df.loc[df[col] == filt]
    return df



from pathlib import Path

import dash_html_components as html

from .. import WebvizContainerABC
from ..webviz_assets import WEBVIZ_ASSETS


class ExampleAssets(WebvizContainerABC):
    def __init__(self, picture_path: Path):
        self.asset_url = WEBVIZ_ASSETS.add(picture_path)

    @property
    def layout(self):
        return html.Img(src=self.asset_url)



import os
import glob

from webviz_config import WebvizConfigTheme

default_theme = WebvizConfigTheme(theme_name="default")  # pylint: disable=invalid-name

default_theme.assets = glob.glob(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "default_assets", "*")
)

default_theme.plotly_layout = {
    "paper_bgcolor": "rgba(90, 90, 90)",
    "plot_bgcolor": "rgba(90, 90, 90)",
    "colorway": [
        "#14213d",
        "#3a2d58",
        "#69356a",
        "#9a3a6f",
        "#c84367",
        "#ea5954",
        "#fe7c37",
        "#ffa600",
    ],
}



import re
import sys
import pathlib
import inspect
import importlib

import yaml

from . import containers as standard_containers
from . import WebvizContainerABC
from .utils import terminal_colors

SPECIAL_ARGS = ["self", "app", "container_settings", "_call_signature", "_imports"]


def _get_webviz_containers(module):
    """Returns a list of all Webviz Containers
    in the module given as input.
    """

    def _is_webviz_container(obj):
        return inspect.isclass(obj) and issubclass(obj, WebvizContainerABC)

    return [member[0] for member in inspect.getmembers(module, _is_webviz_container)]


def _call_signature(
    module,
    module_name,
    container_name,
    container_settings,
    kwargs,
    config_folder,
    contact_person=None,
):
    # pylint: disable=too-many-branches
    """Takes as input the name of a container, the module it is located in,
    together with user given arguments (originating from the configuration
    file). Returns the equivalent Python code wrt. initiating an instance of
    that container (with the given arguments).

    Raises ParserError in the following scenarios:
      * User is missing a required (i.e. no default value) __init__ argument
      * User provided an argument not existing in the class' __init__ function
      * User has given one of the protected arguments in SPECIAL_ARGS
      * If there is type mismatch between user given argument value, and type
        hint in __init__ signature (given that type hint exist)
    """
    argspec = inspect.getfullargspec(getattr(module, container_name).__init__)

    if argspec.defaults is not None:
        required_args = argspec.args[: -len(argspec.defaults)]
    else:
        required_args = argspec.args

    for arg in required_args:
        if arg not in SPECIAL_ARGS and arg not in kwargs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"The container `{container_name}` requires "
                f"the argument `{arg}` in your configuration "
                "file."
                f"{terminal_colors.END}"
            )

    for arg in list(kwargs):
        if arg in SPECIAL_ARGS:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                f"Container argument `{arg}` not allowed."
                f"{terminal_colors.END}"
            )

        if arg == "contact_person":
            if not isinstance(kwargs["contact_person"], dict):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"The contact information provided for "
                    f"container `{container_name}` is "
                    f"not a dictionary. "
                    f"{terminal_colors.END}"
                )

            if any(
                key not in ["name", "phone", "email"]
                for key in kwargs["contact_person"]
            ):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"Unrecognized contact information key "
                    f"given to container `{container_name}`."
                    f'Should be "name", "phone" and/or "email".'
                    f"{terminal_colors.END}"
                )

            contact_person = kwargs.pop("contact_person")

        elif arg not in argspec.args:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "Unrecognized argument. The container "
                f"`{container_name}` does not take an "
                f"argument `{arg}`."
                f"{terminal_colors.END}"
            )

        if arg in argspec.annotations:
            expected_type = argspec.annotations[arg]

            if expected_type == pathlib.Path:
                kwargs[arg] = (config_folder / pathlib.Path(kwargs[arg])).resolve()

            if not isinstance(kwargs[arg], expected_type):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"The value provided for argument `{arg}` "
                    f"given to container `{container_name}` is "
                    f"of type `{type(kwargs[arg]).__name__}`. "
                    f"Expected type "
                    f"`{argspec.annotations[arg].__name__}`."
                    f"{terminal_colors.END}"
                )

    special_args = ""
    if "app" in argspec.args:
        special_args += "app=app, "

    if "container_settings" in argspec.args:
        kwargs["container_settings"] = container_settings

    return (
        f"{module_name}.{container_name}({special_args}**{kwargs})",
        f"container_layout(contact_person={contact_person})",
    )


class ParserError(Exception):
    pass


class ConfigParser:

    STANDARD_CONTAINERS = _get_webviz_containers(standard_containers)

    def __init__(self, yaml_file):

        ConfigParser.check_for_tabs_in_file(yaml_file)

        try:
            self._configuration = yaml.safe_load(open(yaml_file, "r"))
        except yaml.MarkedYAMLError as excep:
            extra_info = (
                f"There is something wrong in the configuration file {yaml_file}. "
            )

            if hasattr(excep, "problem_mark"):
                extra_info += (
                    "The typo is probably somewhere around "
                    f"line {excep.problem_mark.line + 1}."
                )

            raise type(excep)(
                f"{excep}. {terminal_colors.RED}{terminal_colors.BOLD}"
                f"{extra_info}{terminal_colors.END}"
            ).with_traceback(sys.exc_info()[2])

        self._config_folder = pathlib.Path(yaml_file).parent
        self._page_ids = []
        self._assets = set()
        self.clean_configuration()

    @staticmethod
    def check_for_tabs_in_file(path):

        with open(path, "r") as filehandle:
            # Create a list with unique entries of line numbers containing tabs
            lines_with_tabs = list(
                dict.fromkeys(
                    [
                        str(i + 1)
                        for i, line in enumerate(filehandle.readlines())
                        if "\t" in line
                    ]
                )
            )

        if lines_with_tabs:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file contains tabs. You should use ordinary spaces "
                "instead. The following lines contain tabs:\n"
                f"{', '.join(lines_with_tabs)}"
                f"{terminal_colors.END}"
            )

    def _generate_page_id(self, title):
        """From the user given title, this function provides a unique
        human readable page id, not already present in self._page_ids
        """

        base_id = re.sub("[^-a-z0-9_]+", "", title.lower().replace(" ", "_"))

        page_id = base_id

        count = 1
        while page_id in self._page_ids:
            count += 1
            page_id = f"{base_id}{count}"

        return page_id

    def clean_configuration(self):
        # pylint: disable=too-many-branches
        """Various cleaning and checks of the raw configuration read
        from the user provided yaml configuration file.
        """

        self.configuration["_imports"] = set()

        if "container_settings" not in self.configuration:
            container_settings = {}
        else:
            container_settings = self.configuration["container_settings"]

        if "title" not in self.configuration:
            self.configuration["title"] = "Webviz - Powered by Dash"

        if "pages" not in self.configuration:
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration file does not have "
                "information regarding which pages to create."
                f"{terminal_colors.END}"
            )

        if not isinstance(self.configuration["pages"], list):
            raise ParserError(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "The configuration input belonging to the "
                "`pages` keyword should be a list."
                f"{terminal_colors.END}"
            )

        for page_number, page in enumerate(self.configuration["pages"]):

            if "title" not in page:
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    f"Page number {page_number + 1} does "
                    "not have the title specified."
                    f"{terminal_colors.END}"
                )

            if "id" not in page:
                page["id"] = self._generate_page_id(page["title"])
            elif page["id"] in self._page_ids:
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    "You have more than one page "
                    "with the same `id`."
                    f"{terminal_colors.END}"
                )

            self._page_ids.append(page["id"])

            if "content" not in page:
                page["content"] = []
            elif not isinstance(page["content"], list):
                raise ParserError(
                    f"{terminal_colors.RED}{terminal_colors.BOLD}"
                    "The content of page number "
                    f"{page_number + 1} should be a list."
                    f"{terminal_colors.END}"
                )

            containers = [e for e in page["content"] if isinstance(e, dict)]

            for container in containers:
                if "container" not in container:
                    raise ParserError(
                        f"{terminal_colors.RED}{terminal_colors.BOLD}"
                        "Argument `container`, stating name of "
                        "the container to include, is required."
                        f"{terminal_colors.END}"
                    )

                kwargs = {**container}
                container_name = kwargs.pop("container")

                if "." not in container_name:
                    if container_name not in ConfigParser.STANDARD_CONTAINERS:
                        raise ParserError(
                            f"{terminal_colors.RED}{terminal_colors.BOLD}"
                            "You have included an container with "
                            f"name `{container_name}` in your "
                            "configuration file. This is not a "
                            "standard container."
                            f"{terminal_colors.END}"
                        )

                    self.configuration["_imports"].add(
                        ("webviz_config.containers", "standard_containers")
                    )

                    container["_call_signature"] = _call_signature(
                        standard_containers,
                        "standard_containers",
                        container_name,
                        container_settings,
                        kwargs,
                        self._config_folder,
                    )

                    self.assets.update(
                        getattr(standard_containers, container_name).ASSETS
                    )

                else:
                    parts = container_name.split(".")

                    container_name = parts[-1]
                    module_name = ".".join(parts[:-1])
                    module = importlib.import_module(module_name)

                    if container_name not in _get_webviz_containers(module):
                        raise ParserError(
                            f"{terminal_colors.RED}{terminal_colors.BOLD}"
                            f"Module `{module}` does not have a "
                            f"container named `{container_name}`"
                            f"{terminal_colors.END}"
                        )

                    self.configuration["_imports"].add(module_name)
                    container["_call_signature"] = _call_signature(
                        module,
                        module_name,
                        container_name,
                        container_settings,
                        kwargs,
                        self._config_folder,
                    )

                    self.assets.update(getattr(module, container_name).ASSETS)

    @property
    def configuration(self):
        return self._configuration

    @property
    def assets(self):
        return self._assets



import os
import time
import shutil
import tempfile
import subprocess  # nosec

from yaml import YAMLError

from ._config_parser import ParserError
from ._write_script import write_script
from .themes import installed_themes
from .utils import terminal_colors

BUILD_FILENAME = "webviz_app.py"
STATIC_FOLDER = os.path.join(os.path.dirname(__file__), "static")


def build_webviz(args):

    if args.theme not in installed_themes:
        raise ValueError(f"Theme `{args.theme}` is not installed.")

    if args.portable is None:
        build_directory = tempfile.mkdtemp()
    else:
        build_directory = args.portable
        os.makedirs(build_directory)

    shutil.copytree(
        os.path.join(STATIC_FOLDER, "assets"), os.path.join(build_directory, "assets")
    )

    for filename in ["README.md", "Dockerfile", ".dockerignore"]:
        shutil.copy(os.path.join(STATIC_FOLDER, filename), build_directory)

    for asset in installed_themes[args.theme].assets:
        shutil.copy(asset, os.path.join(build_directory, "assets"))

    try:
        if args.portable:
            print(
                f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                "Saving requested data to build folder "
                "such that the webviz instance is portable."
                f"{terminal_colors.END}"
            )

            write_script(
                args, build_directory, "copy_data_template.py.jinja2", "copy_data.py"
            )

            if subprocess.call(  # nosec
                ["python3", "copy_data.py"], cwd=build_directory
            ):
                raise RuntimeError(
                    "Something went wrong. This is probably "
                    "not related to webviz-config, but more "
                    "likely in one of the  dependencies used "
                    "by one or more containers. See exact "
                    "error message and traceback above"
                )

            os.remove(os.path.join(build_directory, "copy_data.py"))

            print(
                f"{terminal_colors.GREEN}{terminal_colors.BOLD}"
                "Finished data extraction. All output saved."
                f"{terminal_colors.END}"
            )

        non_default_assets = write_script(
            args, build_directory, "webviz_template.py.jinja2", BUILD_FILENAME
        )

        for asset in non_default_assets:
            shutil.copy(asset, os.path.join(build_directory, "assets"))

        if not args.portable:
            run_webviz(args, build_directory)

    finally:
        if not args.portable:
            print(f"Deleting temporary folder {build_directory}")
            shutil.rmtree(build_directory)


def run_webviz(args, build_directory):

    print(
        f"{terminal_colors.YELLOW}"
        " Starting up your webviz application. Please wait..."
        f"{terminal_colors.END}"
    )

    app_process = subprocess.Popen(  # nosec
        ["python3", BUILD_FILENAME], cwd=build_directory
    )

    lastmtime = os.path.getmtime(args.yaml_file)

    while app_process.poll() is None:
        time.sleep(1)

        try:
            if lastmtime != os.path.getmtime(args.yaml_file):
                lastmtime = os.path.getmtime(args.yaml_file)
                write_script(
                    args, build_directory, "webviz_template.py.jinja2", BUILD_FILENAME
                )
                print(
                    f"{terminal_colors.BLUE}{terminal_colors.BOLD}"
                    " Rebuilt webviz dash app from configuration file"
                    f"{terminal_colors.END}"
                )

        except (ParserError, YAMLError) as excep:
            print(
                f"{excep} {terminal_colors.RED}{terminal_colors.BOLD}"
                "Fix the error and save the configuration file in "
                " order to trigger a new rebuild."
                f"{terminal_colors.END}"
            )

        except Exception as excep:
            app_process.kill()
            print(
                f"{terminal_colors.RED}{terminal_colors.BOLD}"
                "Unexpected error. Killing the webviz dash application process."
                f"{terminal_colors.END}"
            )
            raise excep
